package db.service;

import db.Globals;
import db.entity.WalItem;
import db.util.FileUtils;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.PreDestroy;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

@Service
@Data
@Slf4j
public class WalFileService {

    @Value("${system.wal-file.prefix}")
    private String prefix;
    @Value("${system.wal-file.suffix}")
    private String suffix;

    ThreadLocal<String> appendableFileName = new ThreadLocal<>();
    ThreadLocal<Long> appendableWalId = new ThreadLocal<>();
    ThreadLocal<FileOutputStream> walWriter = new ThreadLocal<>();


    // 追加写入

    public Long getAppendableWalId() {
        return appendableWalId.get();
    }


    /**
     * 打开新的 wal 文件并创建 writer
     */
    public void openWalFile() {
        appendableFileName.set(prefix + appendableWalId.get() + suffix);
        walWriter.set(FileUtils.openFile(appendableFileName.get()));
        log.info("打开wal文件：{}作为wal写入目标文件", appendableFileName.get());
    }

    /**
     * 打开指定 walId 的 wal 文件并创建 writer
     */
    public void openWalFile(long walId) {
        appendableWalId.set(walId);
        openWalFile();
    }

    @PreDestroy
    public void closeWalFile() {
        if (walWriter.get() == null) {
            return;
        }
        try {
            walWriter.get().close();
            walWriter.set(null);
            log.info("关闭wal文件：{}", appendableFileName.get());
        } catch (IOException e) {
            log.error("无法关闭wal文件：{}", appendableFileName.get(), e);
            walWriter.set(null);
        }
    }

    public void appendWal(WalItem walItem) throws IOException {
        // 如果未打开wal文件，则打开
        if (appendableWalId.get() == null || walWriter.get() == null) {
            appendableWalId.set(System.currentTimeMillis());
            openWalFile();
        }

        // 写入数据
        byte[] walItemPayload = walItem.getPayload();

        walWriter.get().write(walItemPayload);

        log.debug("将{}写入wal文件{}", walItem, appendableFileName.get());
    }

    /**
     * 获取现存所有walId
     *
     * @return
     * @throws IOException
     */
    public List<Long> getOldWalIds() throws IOException {
        List<String> fileNames = FileUtils.listFile(prefix);
        List<Long> walLists = new ArrayList<>();

        for (String fileName : fileNames) {
            if (!fileName.endsWith(suffix)) {
                continue;
            }
            walLists.add(Long.valueOf(fileName.replace(prefix, "").replace(suffix, "")));
        }
        return walLists;
    }

    public List<WalItem> readWal(Long walId) {
        return readWal(prefix + walId + suffix);
    }

    /**
     * 读取指定 wal 文件
     *
     * @return
     * @throws IOException
     */
    public List<WalItem> readWal(String fileName) {
        List<WalItem> walItems = new ArrayList<>();

        int fileLength = FileUtils.fileLength(fileName);
        if (fileLength == 0) {
            return walItems;
        }
        byte[] buffer = new byte[fileLength];
        try {
            FileUtils.readFile(fileName, buffer, 0);
        } catch (IOException e) {
            log.error("读取wal文件失败，filename={}", fileName, e);
            return walItems;
        }

        int offset = 0;
        while (offset < fileLength) {
            WalItem walItem = new WalItem();
            int readLength = walItem.setContent(buffer, offset);
            if (readLength == -1) {
                log.error("walItem文件完整性校验失败，文件名：{}，文件偏移量：{}", fileName, offset);
                return walItems;
            }
            offset += readLength;
            walItems.add(walItem);
        }

        log.info("读取wal文件成功: {}", fileName);
        return walItems;
    }

    /**
     * 清空正在写入的wal文件
     */
    public void clear() {
        closeWalFile();
        FileUtils.deleteFile(appendableFileName.get());
        log.info("清空wal文件{}", appendableFileName.get());
    }

    /**
     * 清空指定id的wal文件
     *
     * @param walId
     */
    public void clear(long walId) {
        String fileName = prefix + walId + suffix;
        FileUtils.deleteFile(fileName);
        log.info("清空wal文件{}", fileName);
    }
}
