package db.netty;

import db.entity.packages.BasePackage;
import io.netty.bootstrap.Bootstrap;
import io.netty.channel.*;
import io.netty.channel.epoll.Epoll;
import io.netty.channel.epoll.EpollChannelOption;
import io.netty.channel.epoll.EpollDatagramChannel;
import io.netty.channel.epoll.EpollEventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioDatagramChannel;
import io.netty.util.concurrent.DefaultEventExecutorGroup;
import io.netty.util.concurrent.EventExecutorGroup;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.DependsOn;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.util.concurrent.DelayQueue;

@Component
@DependsOn("springUtils")
@Slf4j
public class NettyServer {

    EventLoopGroup bossGroup;

    EventExecutorGroup workerGroup;

    @Value("${netty.port}")
    private Integer port;
    @Value("${netty.host}")
    private String host;

    @Value("${logging.showRecvBytes}")
    private Boolean showRecvBytes;

    private Channel channel;

    @Value("${system.sender.retry-stamp}")
    private Integer retryStamp;
    @Value("${system.sender.retry-times}")
    private Integer retryTimes;

    @Value("${system.udp-server.boss-thread-num}")
    private Integer bossThreadNum;
    @Value("${system.udp-server.worker-thread-num}")
    private Integer workerThreadNum;

    @Value("${system.udp-server.use-multithread}")
    private Boolean useMultiThread;

    @Value("${debug.justReceivePacket}")
    private Boolean isJustReceivePacket;
    @Value("${debug.decode}")
    private Boolean decodePacketOrNot;

    @Value("${netty.recvBufferSize}")
    private Integer recvBufferSize;

    private DelayQueue<BasePackage> sendingQueue = new DelayQueue<>();

    /**
     * 启动Netty Server
     *
     * @throws InterruptedException
     */
    @PostConstruct
    public void start() throws InterruptedException {
        useMultiThread = useMultiThread && Epoll.isAvailable(); // 必须系统支持

        bossGroup = useMultiThread ? new EpollEventLoopGroup() : new NioEventLoopGroup();

        workerGroup = new DefaultEventExecutorGroup(workerThreadNum);//业务线程池

        log.info("开始启动 Netty Server，host:{} port: {}", host, port);
        log.info("***Netty-Server isJustReceivePacket:{}", isJustReceivePacket);
        log.info("***Netty-Server recvBufferSize:{}", recvBufferSize);
        Bootstrap bootstrap = new Bootstrap();

        Encoder encoder = new Encoder();
        Decoder decoder = new Decoder();
        NettyServerHandler nettyServerHandler = new NettyServerHandler();
        decoder.setShowRecvBytes(this.showRecvBytes);
        decoder.setDecode(this.decodePacketOrNot);
        log.info("Netty-Server decode showRecvBytes:{} decodePacketOrNot:{}", this.showRecvBytes, this.decodePacketOrNot);

        bootstrap.group(bossGroup)
                // 指定Channel
                .channel(useMultiThread ? EpollDatagramChannel.class : NioDatagramChannel.class)

                .option(ChannelOption.SO_BROADCAST, true)
                .option(ChannelOption.SO_RCVBUF, 1024 * 1024 * recvBufferSize)

                .handler(new ChannelInitializer<Channel>() {
                    @Override
                    public void initChannel(final Channel ch) throws Exception {
                        ChannelPipeline p = ch.pipeline();
                        p.addLast(workerGroup, encoder, decoder);
                        p.addLast(workerGroup, nettyServerHandler);
                    }
                });
        if (useMultiThread) {
            bootstrap.option(EpollChannelOption.SO_REUSEPORT, true);
        }

        if (useMultiThread) {
            // linux系统下使用SO_REUSEPORT特性，使得多个线程绑定同一个端口
            if (bossThreadNum <= 0) {
                bossThreadNum = Runtime.getRuntime().availableProcessors();
            }
            log.info("使用 epoll reuseport，启动线程数：{}", bossThreadNum);
            for (int i = 0; i < bossThreadNum; i++) {
                ChannelFuture future = bootstrap.bind(host, port).await();
                if (!future.isSuccess()) {
                    log.error("[多线程-{}]失败启动 Netty Server，host:{} port: {}", i, host, port);
                } else {
                    log.info("[多线程-{}]成功启动 Netty Server，host:{} port: {}", i, host, port);
                    channel = future.channel();
                }
            }
        } else {
            ChannelFuture future = bootstrap.bind(host, port).sync();
            if (future.isSuccess()) {
                log.info("[单线程]成功启动 Netty Server，host:{} port: {}", host, port);
                channel = future.channel();
            } else {
                log.error("[单线程]失败启动 Netty Server，host:{} port: {}", host, port);
            }
        }
    }

    @PostConstruct
    public void startCheckSendingQueue() {
        int i = 0;
        new Thread(() -> {
            checkSendingQueue();
        }, "send-queue-" + i).start();
    }

    private void checkSendingQueue() {
        while (true) {
            try {
                BasePackage basePackage = sendingQueue.take();
                log.info("发出{}", basePackage);
                channel.writeAndFlush(basePackage);
                if (basePackage.getSendTimes() > 1) {
                    // 再次放入队列
                    basePackage.setSendTimes(basePackage.getSendTimes() - 1);
                    basePackage.setSendStamp(System.currentTimeMillis() + retryStamp);
                    send(basePackage);
                    log.info("超时重传机制，重新添加至发送队列{}", basePackage);
                }
            } catch (InterruptedException e) {
                log.error("从发送队列队列中取出报文失败", e);
            }
        }
    }

    public void send(BasePackage basePackage) {
        // 放入发送队列
        if (basePackage.getSendStamp() == null) {
            // 首次发送，添加时间
            basePackage.setSendTimes(1);
            basePackage.setSendStamp(System.currentTimeMillis());
        }
        log.info("添加消息至发送队列{}", basePackage);
        sendingQueue.put(basePackage);
    }

    @PreDestroy
    public void destroy() {
        bossGroup.shutdownGracefully();
        log.info("关闭Netty");
    }
}
