package db.service;

import db.DBApiEntry;
import db.DBVal;
import db.Point;
import db.entity.LogFileMeta;
import db.mapper.LogFileMetaMapper;
import db.util.ConvertFormat;
import db.util.SpentTimeCalculator;
import db.util.Util;
import db.utils.PcapHelper;
import lombok.extern.slf4j.Slf4j;
import org.pcap4j.core.*;
import org.pcap4j.packet.UnknownPacket;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Service
@Slf4j
public class QueryService {

    @Autowired
    private LogFileService logFileService;

    @Autowired
    private LogFileMetaMapper logFileMetaMapper;

    @Value("${reso-db.host}")
    private String host;

    @Value("${reso-db.port}")
    private int port;

    @Value("${reso-db.metric-name}")
    private String metricName;

    @Value("${server.port}")
    private String serverPort;

    @Value("${ip.host}")
    private String hostIP;

    @Value("${system.pcap-file.prefix}")
    private String pcapPrefix;

    @Value("${system.pcap-file.suffix}")
    private String pcapSuffix;

    @Value("${system.pcap-file.template}")
    private String pcapTemplate;

    private DBApiEntry entry;

    @PostConstruct
    public void afterCons() {
        this.entry = DBApiEntry.initApiEntry(host, port);
    }

    public String queryByTimeWithCondition(String start, String end, Map<String, String> tags) throws ParseException,
            IOException, PcapNativeException, NotOpenException {

        Point point = new Point(metricName, tags);

        long startTime = Util.dateStringToUTCMilliSeconds(start);
        long endTime = Util.dateStringToUTCMilliSeconds(end);

        DBApiEntry entry = DBApiEntry.initApiEntry(host, port);

        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create();
        spentTimeCalculator.begin();
        List<DBVal> lists = entry.getHistRaw(point, startTime, endTime);
        spentTimeCalculator.end();
        log.info("queryByTimeAndCondition，查询条件为：start:{}, end:{}, point:{}", startTime, endTime, point);
        log.info("查询ResoDB，查询到条目：{}，查询耗时：{}", lists.size(), spentTimeCalculator.getSpendTime());

        spentTimeCalculator.begin();
        HashMap<Long, ArrayList<Integer>> map = new ConvertFormat().convertToLongAndGetIndex(lists);
        spentTimeCalculator.end();
        log.info("记录分片耗时：{}", spentTimeCalculator.getSpendTime());

        ArrayList<byte[]> logItems = new ArrayList<>();
        for (Long key : map.keySet()) {
            logItems.addAll(logFileService.querySingleLogFile(key, map.get(key)));
        }

        spentTimeCalculator.begin();
        PcapHandle handle = Pcaps.openOffline(pcapTemplate);
        String pcapName = new StringBuilder().append(System.currentTimeMillis())
                .append(startTime).append(endTime).toString();
        PcapDumper dumper = handle.dumpOpen(pcapPrefix + pcapName + pcapSuffix);

        for (byte[] logItem : logItems) {
            UnknownPacket packet = UnknownPacket.newPacket(logItem, 0, logItem.length);
            dumper.dump(packet);
        }
        spentTimeCalculator.end();
        log.info("生成Pcap文件时间：{}", spentTimeCalculator.getSpendTime());

        dumper.close();
        handle.close();

        StringBuilder returnLink = new StringBuilder();
        returnLink.append("http://").append(hostIP).append(":")
                .append(serverPort).append("/download?pcapName=").append(pcapName);

        log.info("返回链接：{}", returnLink);
        return String.valueOf(returnLink);
    }

    public String queryByTime(String start, String end) throws ParseException, IOException {

        long startTime = Util.dateStringToUTCMilliSeconds(start);
        long endTime = Util.dateStringToUTCMilliSeconds(end);

        log.debug("queryByTime，查询条件为：start:{}, end:{}", startTime, endTime);

        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create();
        spentTimeCalculator.begin();

        List<LogFileMeta> logFileMetas = logFileMetaMapper.getValidLogFileMetaBetweenTime(startTime, endTime);

        spentTimeCalculator.end();
        log.debug("queryByTime，查询sqlite，查询条目：{}，查询时间：{}", logFileMetas.size(), spentTimeCalculator.getSpendTime());

        String pcapName = new StringBuilder().append(System.currentTimeMillis())
                .append(startTime).append(endTime).toString();

        spentTimeCalculator.begin();
        PcapHelper pcapHelper = new PcapHelper();
        for (LogFileMeta logFileMeta : logFileMetas) {
            List<byte[]> bytes = logFileService.queryLogBetweenTimeRange(logFileMeta, startTime, endTime);
            for (byte[] item : bytes) {
                pcapHelper.addItem(item.length, item);
            }
        }

        try (FileOutputStream fileOutputStream = new FileOutputStream(pcapPrefix + pcapName + pcapSuffix)) {
            byte[] bytes = pcapHelper.getBytes();
            fileOutputStream.write(bytes, 0, bytes.length);
            fileOutputStream.flush();
        } catch (IOException e) {
            log.info("生成pcap文件失败:{}",e);
        }

        spentTimeCalculator.end();
        log.info("queryByTime，查询磁盘并生成pcap总时间：{}", spentTimeCalculator.getSpendTime());

        StringBuilder returnLink = new StringBuilder();
        returnLink.append("http://").append(hostIP).append(":")
                .append(serverPort).append("/download?pcapName=").append(pcapName);

        log.info("返回链接：{}", returnLink);
        return String.valueOf(returnLink);
    }

}
